"""This module contains the public api of the fastrlapi reinforcement learning library.

    It consist mainly of the class hierarchy of the available agents (algorithms), registrations and
    the management of the available backends. In their implementation the agents forward their calls
    to the chosen backend.
"""

from abc import ABC
from collections import namedtuple
import json
import os
import statistics
from typing import Dict, List, Optional, Tuple, Type, Union

from fastrlapi import core
from fastrlapi.backends import core as bcore
from fastrlapi.callbacks import plot
import fastrlapi.backends.default
import fastrlapi.backends.tfagents
import tensorflow as tf

# import fastrlapi.backends.tforce

_backends: [bcore.BackendAgentFactory] = []

"""The seed used for all agents and gym environments. If None no seed is set (default)."""
seed: Optional[int] = None


def load(directory: str,
         callbacks: Union[List[core.AgentCallback], core.AgentCallback, None] = None):
    """Loads an agent from directory.

    After a successful load play() may be called directly.
    The agent, model, backend, seed and play policy are restored according to the previously saved agent.

    Args:
        directory: the directory containing the previously saved policy.
        callbacks: list of callbacks called during save (eg log.Agent)

    Result:
        a new instance of fastrlapi
    """
    assert directory
    agent_json_path = os.path.join(directory, EasyAgent._KEY_EASYAGENT_FILENAME)
    policy_directory = os.path.join(directory, EasyAgent._KEY_POLICY_DIRECTORY)
    assert os.path.isdir(directory), f'directory "{directory}" not found.'
    assert os.path.isfile(agent_json_path), 'file "{agent_json_path}" not found.'
    assert os.path.isdir(policy_directory), f'directory "{policy_directory}" not found.'
    with open(agent_json_path) as jsonfile:
        agent_dict = json.load(jsonfile)
    result = EasyAgent._from_dict(agent_dict)
    callbacks = result._to_callback_list(callbacks=callbacks)
    result._backend_agent.load(directory=policy_directory, callbacks=callbacks)
    return result


def register_backend(backend: bcore.BackendAgentFactory):
    """registers a backend as a factory for agent implementations.

    If another backend with the same name is already registered, the old backend is replaced by backend.
    """
    assert backend
    old_backends = [b for b in _backends if b.backend_name == backend.backend_name]
    for old_backend in old_backends:
        _backends.remove(old_backend)
    _backends.append(backend)


def activate_tensorforce():
    """registers the tensorforce backend.

    Due to an incompatibility between tensorforce and tf-agents, both libraries may not run
    in the same python instance. Thus - for the time being - once this method is called,
    the tfagents backend may not be used anymore.
    """
    import fastrlapi.backends.tforce

    global _backends

    assert  fastrlapi.backends.core._tf_eager_execution_active is None or \
            fastrlapi.backends.core._tf_eager_execution_active == False, \
            "tensorforce can not be activated, since tensorflow eager execution mode was already actived."

    _backends = []
    register_backend(fastrlapi.backends.default.DefaultAgentFactory(register_tensorforce=True))
    register_backend(fastrlapi.backends.tforce.TensorforceAgentFactory())

def _activate_tfagents():
    """registers the tfagents backend.

    Due to an incompatibility between tensorforce and tf-agents, both libraries may not run
    in the same python instance.
    """
    global _backends

    assert  fastrlapi.backends.core._tf_eager_execution_active is None or \
            fastrlapi.backends.core._tf_eager_execution_active == True, \
            "tfagents can not be activated, since tensorflow eager execution mode was already disabled."
    _backends = []
    register_backend(fastrlapi.backends.default.DefaultAgentFactory(register_tensorforce=False))
    register_backend(fastrlapi.backends.tfagents.TfAgentAgentFactory())


_activate_tfagents()


class EasyAgent(ABC):
    """Abstract base class for all easy reinforcment learning agents.

    Besides forwarding train and play it implements persistence."""

    _KEY_BACKEND = 'backend'
    _KEY_EASYAGENT_CLASS = 'easyagent_class'
    _KEY_EASYAGENT_FILENAME = 'easyagent.json'
    _KEY_MODEL_CONFIG = 'model_config'
    _KEY_POLICY_DIRECTORY = 'policy'
    _KEY_VERSION = 'version'

    def __init__(self,
                 gym_env_name: str,
                 fc_layers: Union[Tuple[int, ...], int, None] = None,
                 backend: str = None):
        """
            Args:
                gym_env_name: name of an OpenAI gym environment to be used for training and evaluation
                fc_layers: defines the neural network to be used, a sequence of fully connected
                    layers of the given size. Eg (75,40) yields a neural network consisting
                    out of 2 hidden layers, the first one containing 75 and the second layer
                    containing 40 neurons.
                backend=the backend to be used (eg 'tfagents'), if None a default implementation is used.
                    call get_backends() to get a list of the available backends.
        """
        model_config = core.ModelConfig(gym_env_name=gym_env_name, fc_layers=fc_layers, seed=seed)
        self._initialize(model_config=model_config, backend_name=backend)
        return

    def _initialize(self, model_config: core.ModelConfig, backend_name: str = None):
        if backend_name is None:
            backend_name = fastrlapi.backends.default.DefaultAgentFactory.backend_name
        backend: bcore.BackendAgentFactory = _get_backend(backend_name)

        assert model_config is not None, "model_config not set."
        assert backend, f'Backend "{backend_name}" not found. The registered backends are {get_backends()}.'

        self._model_config: core.ModelConfig = model_config
        backend_agent = backend.create_agent(easyagent_type=type(self), model_config=model_config)
        assert backend_agent, f'Backend "{backend_name}" does not implement "{type(self).__name__}". ' + \
                              f'Choose one of the following backend {get_backends(type(self))}.'
        self._backend_agent: Optional[bcore._BackendAgent] = backend_agent
        self._backend_name: str = backend_name
        self._backend_agent._agent_context._agent_saver = self.save
        return

    def _add_plot_callbacks(self, callbacks: List[core.AgentCallback],
                            default_plots: Optional[bool],
                            default_plot_callbacks: List[plot._PlotCallback]
                            ) -> List[core.AgentCallback]:
        """Adds the default callbacks and sorts all callbacks in the order
            _PreProcessCallbacks, AgentCallbacks, _PostProcessCallbacks.

        Args:
            callbacks: existing callbacks to prepare
            default_plots: if set or if None and callbacks does not contain plots then the default plots are added
            default_plot_callbacks: plot callbacks to add.
        """
        pre_process: List[core.AgentCallback] = [plot._PreProcess()]
        agent: List[core.AgentCallback] = []
        post_process: List[core.AgentCallback] = [plot._PostProcess()]

        if default_plots is None:
            default_plots = True
            for c in callbacks:
                default_plots = default_plots and (not isinstance(c, plot._PlotCallback))
        if default_plots:
            agent = default_plot_callbacks

        for c in callbacks:
            if isinstance(c, core._PreProcessCallback):
                pre_process.append(c)
            else:
                if isinstance(c, core._PostProcessCallback):
                    post_process.append(c)
                else:
                    agent.append(c)
        result: List[core.AgentCallback] = pre_process + agent + post_process
        return result

    @staticmethod
    def _from_dict(param_dict: Dict[str, object]):
        """recreates a new agent instance according to the definition previously created by _to_dict.

        Returns:
            new agent instance (excluding any trained policy), the agent type is preserved.
        """
        assert param_dict
        mc: core.ModelConfig = core.ModelConfig._from_dict(param_dict[EasyAgent._KEY_MODEL_CONFIG])
        agent_class = globals()[param_dict[EasyAgent._KEY_EASYAGENT_CLASS]]
        backend: str = param_dict[EasyAgent._KEY_BACKEND]
        result = agent_class(gym_env_name=mc.original_env_name, backend=backend)
        result._initialize(model_config=mc, backend_name=backend)
        return result

    def _to_callback_list(self, callbacks: Union[Optional[core.AgentCallback], List[core.AgentCallback]]) -> List[
        core.AgentCallback]:
        """maps callbacks to an admissible callback list.

        if callbacks is None an empty list is returned.
        if callbacks is an AgentCallback a list containing only this callback is returned
        otherwise callbacks is returned
        """
        result: List[core.AgentCallback] = []
        if not callbacks is None:
            if isinstance(callbacks, core.AgentCallback):
                result = [callbacks]
            else:
                assert isinstance(callbacks, list), "callback not an AgentCallback or a list thereof."
                result = callbacks
        return result

    def _to_dict(self) -> Dict[str, object]:
        """saves the agent definition to a dict.

            Returns:
                dict containing all parameters to recreate the agent (excluding a trained policy)
        """
        result: Dict[str, object] = dict()
        result[EasyAgent._KEY_VERSION] = fastrlapi.__version__
        result[EasyAgent._KEY_EASYAGENT_CLASS] = self.__class__.__name__
        result[EasyAgent._KEY_BACKEND] = self._backend_name
        result[EasyAgent._KEY_MODEL_CONFIG] = self._model_config._to_dict()
        result[EasyAgent._KEY_POLICY_DIRECTORY] = 'policy'
        return result

    def evaluate(self,
                 num_episodes: int = 50,
                 max_steps_per_episode: int = 50):
        """Plays num_episodes with the current policy and computes metrics on rewards.

        Args:
            num_episodes: number of episodes to play
            max_steps_per_episode: max steps per episode

        Returns:
            extensible score metrics
        """
        play_context = core.PlayContext()
        play_context.max_steps_per_episode = max_steps_per_episode
        play_context.num_episodes = num_episodes
        self.play(play_context=play_context, default_plots=False)
        Metrics = namedtuple('Metrics', 'steps rewards')

        Rewards = namedtuple('Rewards', 'mean std min max all')
        all_rewards = list(play_context.sum_of_rewards.values())
        mean_reward, std_reward, min_reward, max_reward = statistics.mean(all_rewards), statistics.stdev(
            all_rewards), min(all_rewards), max(all_rewards)
        rewards = Rewards(mean=mean_reward, std=std_reward, min=min_reward, max=max_reward, all=all_rewards)

        Steps = namedtuple('Steps', 'mean std min max all')
        all_num_steps = []
        for i in play_context.rewards.keys():
            all_num_steps.append(len(play_context.rewards[i]))

        mean_steps, std_steps, min_steps, max_steps = statistics.mean(all_num_steps), statistics.stdev(
            all_num_steps), min(all_num_steps), max(all_num_steps)
        steps = Steps(mean=mean_steps, std=std_steps, min=min_steps, max=max_steps, all=all_num_steps)

        metrics = Metrics(rewards=rewards, steps=steps)
        return metrics

    def play(self,
             callbacks: Union[List[core.AgentCallback], core.AgentCallback, None] = None,
             num_episodes: int = 1,
             max_steps_per_episode: int = 1000,
             play_context: core.PlayContext = None,
             default_plots: bool = None):
        """Plays num_episodes with the current policy.

        Args:
            callbacks: list of callbacks called during each episode play
            num_episodes: number of episodes to play
            max_steps_per_episode: max steps per episode
            play_context: play configuration to be used. If set override all other play context arguments
            default_plots: if set addes a set of default callbacks (plot.State, plot.Rewards, ...)

        Returns:
            play_context containg the actions taken and the rewards received during training
        """
        assert self._backend_agent._agent_context._is_policy_trained, "No trained policy available. Call train() first."
        if play_context is None:
            play_context = core.PlayContext()
            play_context.max_steps_per_episode = max_steps_per_episode
            play_context.num_episodes = num_episodes
        callbacks = self._to_callback_list(callbacks=callbacks)
        callbacks = self._add_plot_callbacks(callbacks, default_plots, [plot.Steps(), plot.Rewards()])
        self._backend_agent.play(play_context=play_context, callbacks=callbacks)
        return play_context

    def save(self, directory: Optional[str] = None,
             callbacks: Union[List[core.AgentCallback], core.AgentCallback, None] = None) -> str:
        """Saves the currently trained actor policy in directory.

        If save is called before a trained policy is created, eg by calling train, an exception is raised.

        Args:
             directory: the directory to save the policy weights to.
                if the directory does not exist yet, a new directory is created. if None the policy is saved
                in a temp directory.
             callbacks: list of callbacks called during save (eg log.Agent)

        Returns:
            the absolute path to the directory containing the saved policy.
        """
        if directory is None:
            directory = bcore._get_temp_path()
        assert directory
        assert self._backend_agent._agent_context._is_policy_trained, "No trained policy available. Call train() first."

        directory = bcore._mkdir(directory)
        agent_json_path = os.path.join(directory, EasyAgent._KEY_EASYAGENT_FILENAME)
        with open(agent_json_path, 'w') as jsonfile:
            agent_dict = self._to_dict()
            json.dump(agent_dict, jsonfile, sort_keys=True, indent=2)
        callbacks = self._to_callback_list(callbacks=callbacks)
        policy_directory = os.path.join(directory, EasyAgent._KEY_POLICY_DIRECTORY)
        policy_directory = bcore._mkdir(policy_directory)
        self._backend_agent.save(directory=policy_directory, callbacks=callbacks)
        return directory

    def train(self, train_context: core.TrainContext,
              callbacks: Union[List[core.AgentCallback], core.AgentCallback, None],
              default_plots: Optional[bool]):
        """Trains a new model using the gym environment passed during instantiation.

        Args:
            callbacks: list of callbacks called during the training and evaluation
            train_context: training configuration to be used (num_iterations,num_episodes_per_iteration,...)
            default_plots: if set adds a set of default callbacks (plot.State, plot.Rewards, plot.Loss,...).
                if None default callbacks are only added if the callbacks list is empty
        """
        assert train_context, "train_context not set."
        callbacks = self._to_callback_list(callbacks=callbacks)
        callbacks = self._add_plot_callbacks(callbacks, default_plots, [plot.Loss(), plot.Steps(), plot.Rewards()])
        self._backend_agent.train(train_context=train_context, callbacks=callbacks)


def get_backends(agent: Optional[Type[EasyAgent]] = None):
    """returns a list of all registered backends containing an implementation for the EasyAgent type agent.

    Args:
        agent: type deriving from EasyAgent for which the backend identifiers are returned.

    Returns:
        a list of admissible values for the 'backend' argument of EazyAgents constructors or a list of all
        available backends if agent is None.
    """
    result = [b.backend_name for b in _backends]
    if agent:
        result = [b.backend_name for b in _backends if agent in b.get_algorithms()]
    return result


def _get_backend(backend_name: str):
    """Yields the backend with the given name.

    Returns:
        the backend instance or None if no backend is found."""
    assert backend_name
    backends = [b for b in _backends if b.backend_name == backend_name]
    assert len(backends) <= 1, f'no backend found with name "{backend_name}". Available backends = {get_backends()}'
    result = None
    if backends:
        result = backends[0]
    return result


class CemAgent(EasyAgent):
    """creates a new agent based on the cross-entropy-method algorithm.

       From https://learning.mpi-sws.org/mlss2016/slides/2016-MLSS-RL.pdf:
        Initialize µ ∈Rd,σ ∈Rd
        for iteration = 1,2,... num_iterations do
            Collect num_episodes_per_iteration samples of θi ∼ N(µ,diag(σ))
            Perform a noisy evaluation Ri ∼ θi
            Select the top elite_set_fraction of samples (e.g. p = 0.2), which we’ll call the elite set
            Fit a Gaussian distribution, with diagonal covariance, to the elite set, obtaining a new µ,σ.
        end for
        Return the ﬁnal µ.

        see https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.81.6579&rep=rep1&type=pdf
    """

    def __init__(self, gym_env_name: str, fc_layers: Optional[Tuple[int, ...]] = None, backend: str = None):
        super().__init__(gym_env_name, fc_layers, backend)
        assert False, "CemAgent is currently not available (pending migration of keras-rl to tf2.0)"

    def train(self,
              callbacks: Union[List[core.AgentCallback], core.AgentCallback, None] = None,
              num_iterations: int = 100,
              num_episodes_per_iteration: int = 50,
              max_steps_per_episode: int = 500,
              elite_set_fraction: float = 0.1,
              num_iterations_between_eval: int = 5,
              num_episodes_per_eval: int = 10,
              train_context: core.CemTrainContext = None,
              default_plots: bool = None):
        """Trains a new model using the gym environment passed during instantiation.

        Args:
            callbacks: list of callbacks called during training and evaluation
            num_iterations: number of times the training is repeated (with additional data)
            num_episodes_per_iteration: number of episodes played in each iteration. for each episode a new
                policy is sampled from the current weight distribution.
            max_steps_per_episode: maximum number of steps per episode
            elite_set_fraction: the fraction of policies which are members of the elite set.
                These policies are used to fit a new weight distribution in each iteration.
            num_iterations_between_eval: number of training iterations before the current policy is evaluated.
                if 0 no evaluation is performed.
            num_episodes_per_eval: number of episodes played to estimate the average return and steps
            train_context: training configuration to be used. if set overrides all other training context arguments.
            default_plots: if set adds a set of default callbacks (plot.State, plot.Rewards, plot.Loss,...).
                if None default callbacks are only added if the callbacks list is empty

        Returns:
            train_context: the training configuration containing the loss and sum of rewards encountered
                during training
        """
        if train_context is None:
            train_context = core.CemTrainContext()
            train_context.num_iterations = num_iterations
            train_context.max_steps_per_episode = max_steps_per_episode
            train_context.elite_set_fraction = elite_set_fraction
            train_context.num_iterations_between_eval = num_iterations_between_eval
            train_context.num_episodes_per_eval = num_episodes_per_eval

        super().train(train_context=train_context, callbacks=callbacks, default_plots=default_plots)
        return train_context


class DqnAgent(EasyAgent):
    """creates a new agent based on the Dqn algorithm.

    From wikipedia:
    The DeepMind system used a deep convolutional neural network, with layers of tiled convolutional filters to mimic
    the effects of receptive fields. Reinforcement learning is unstable or divergent when a nonlinear function
    approximator such as a neural network is used to represent Q.
    This instability comes from the correlations present in the sequence of observations, the fact that small updates
    to Q may significantly change the policy and the data distribution, and the correlations between Q and the
    target values.

    The technique used experience replay, a biologically inspired mechanism that uses a random sample of prior actions
    instead of the most recent action to proceed.[2] This removes correlations in the observation sequence and smooths
    changes in the data distribution. Iterative update adjusts Q towards target values that are only periodically
    updated, further reducing correlations with the target.[17]

    see also: https://deepmind.com/research/publications/human-level-control-through-deep-reinforcement-learning
    """

    def train(self,
              callbacks: Union[List[core.AgentCallback], core.AgentCallback, None] = None,
              num_iterations: int = 20000,
              max_steps_per_episode: int = 500,
              num_steps_per_iteration: int = 1,
              num_steps_buffer_preload=1000,
              num_steps_sampled_from_buffer=64,
              num_iterations_between_eval: int = 1000,
              num_episodes_per_eval: int = 10,
              learning_rate: float = 0.001,
              train_context: core.StepsTrainContext = None,
              default_plots: bool = None):
        """Trains a new model using the gym environment passed during instantiation.

        Args:
            callbacks: list of callbacks called during training and evaluation
            num_iterations: number of times the training is repeated (with additional data)
            max_steps_per_episode: maximum number of steps per episode
            num_steps_per_iteration: number of steps played per training iteration
            num_steps_buffer_preload: number of initial collect steps to preload the buffer
            num_steps_sampled_from_buffer: the number of steps sampled from buffer for each iteration training
            num_iterations_between_eval: number of training iterations before the current policy is evaluated.
                if 0 no evaluation is performed.
            num_episodes_per_eval: number of episodes played to estimate the average return and steps
            learning_rate: the learning rate used in the next iteration's policy training (0,1]
            train_context: training configuration to be used. if set overrides all other training context arguments.
            default_plots: if set adds a set of default callbacks (plot.State, plot.Rewards, plot.Loss,...).
                if None default callbacks are only added if the callbacks list is empty

        Returns:
            train_context: the training configuration containing the loss and sum of rewards encountered
                during training
        """
        if train_context is None:
            train_context = core.StepsTrainContext()
            train_context.num_iterations = num_iterations
            train_context.max_steps_per_episode = max_steps_per_episode
            train_context.num_steps_per_iteration = num_steps_per_iteration
            train_context.num_steps_buffer_preload = num_steps_buffer_preload
            train_context.num_steps_sampled_from_buffer = num_steps_sampled_from_buffer
            train_context.num_iterations_between_eval = num_iterations_between_eval
            train_context.num_episodes_per_eval = num_episodes_per_eval
            train_context.learning_rate = learning_rate

        super().train(train_context=train_context, callbacks=callbacks, default_plots=default_plots)
        return train_context


class DoubleDqnAgent(DqnAgent):
    """Agent based on the Double Dqn algorithm (https://arxiv.org/abs/1509.06461)"""


class DuelingDqnAgent(DqnAgent):
    """Agent based on the Dueling Dqn algorithm (https://arxiv.org/abs/1511.06581)."""


class PpoAgent(EasyAgent):
    """creates a new agent based on the PPO algorithm.

        PPO is an actor-critic algorithm using 2 neural networks. The actor network
        to predict the next action to be taken and the critic network to estimate
        the value of the game state we are currently in (the expected, discounted
        sum of future rewards when following the current actor network).

        see also: https://spinningup.openai.com/en/latest/algorithms/ppo.html
    """

    def train(self,
              callbacks: Union[List[core.AgentCallback], core.AgentCallback, None] = None,
              num_iterations: int = 100,
              num_episodes_per_iteration: int = 10,
              max_steps_per_episode: int = 500,
              num_epochs_per_iteration: int = 10,
              num_iterations_between_eval: int = 5,
              num_episodes_per_eval: int = 10,
              learning_rate: float = 0.001,
              train_context: core.PpoTrainContext = None,
              default_plots: bool = None):
        """Trains a new model using the gym environment passed during instantiation.

        Args:
            callbacks: list of callbacks called during training and evaluation
            num_iterations: number of times the training is repeated (with additional data)
            num_episodes_per_iteration: number of episodes played per training iteration
            max_steps_per_episode: maximum number of steps per episode
            num_epochs_per_iteration: number of times the data collected for the current iteration
                is used to retrain the current policy
            num_iterations_between_eval: number of training iterations before the current policy is evaluated.
                if 0 no evaluation is performed.
            num_episodes_per_eval: number of episodes played to estimate the average return and steps
            learning_rate: the learning rate used in the next iteration's policy training (0,1]
            train_context: training configuration to be used. if set overrides all other training context arguments.
            default_plots: if set adds a set of default callbacks (plot.State, plot.Rewards, plot.Loss,...).
                if None default callbacks are only added if the callbacks list is empty

        Returns:
            train_context: the training configuration containing the loss and sum of rewards encountered
                during training
        """
        if train_context is None:
            train_context = core.PpoTrainContext()
            train_context.num_iterations = num_iterations
            train_context.num_episodes_per_iteration = num_episodes_per_iteration
            train_context.max_steps_per_episode = max_steps_per_episode
            train_context.num_epochs_per_iteration = num_epochs_per_iteration
            train_context.num_iterations_between_eval = num_iterations_between_eval
            train_context.num_episodes_per_eval = num_episodes_per_eval
            train_context.learning_rate = learning_rate

        super().train(train_context=train_context, callbacks=callbacks, default_plots=default_plots)
        return train_context


class RandomAgent(EasyAgent):
    """Agent which always chooses uniform random actions."""

    def train(self,
              callbacks: Union[List[core.AgentCallback], core.AgentCallback, None] = None,
              num_iterations: int = 10,
              max_steps_per_episode: int = 1000,
              num_episodes_per_eval: int = 10,
              train_context: core.TrainContext = None,
              default_plots: bool = None):
        """Evaluates the environment using a uniform random policy.

        The evaluation is performed in batches of num_episodes_per_eval episodes.

        Args:
            callbacks: list of callbacks called during training and evaluation
            num_iterations: number of times a batch of num_episodes_per_eval episodes is evaluated.
            max_steps_per_episode: maximum number of steps per episode
            num_episodes_per_eval: number of episodes played to estimate the average return and steps
            train_context: training configuration to be used. if set overrides all other training context arguments.
            default_plots: if set adds a set of default callbacks (plot.State, plot.Rewards, plot.Loss,...)

        Returns:
            train_context: the training configuration containing the loss and sum of rewards encountered
                during training
        """
        if train_context is None:
            train_context = core.TrainContext()
            train_context.num_iterations = num_iterations
            train_context.max_steps_per_episode = max_steps_per_episode
            train_context.num_epochs_per_iteration = 0
            train_context.num_iterations_between_eval = 1
            train_context.num_episodes_per_eval = num_episodes_per_eval
            train_context.learning_rate = 1

        super().train(train_context=train_context, callbacks=callbacks, default_plots=default_plots)
        return train_context


class ReinforceAgent(EasyAgent):
    """creates a new agent based on the Reinforce algorithm.

        Reinforce is a vanilla policy gradient algorithm using a single actor network.

        see also: www-anw.cs.umass.edu/~barto/courses/cs687/williams92simple.pdf
    """

    def train(self,
              callbacks: Union[List[core.AgentCallback], core.AgentCallback, None] = None,
              num_iterations: int = 100,
              num_episodes_per_iteration: int = 10,
              max_steps_per_episode: int = 500,
              num_epochs_per_iteration: int = 10,
              num_iterations_between_eval: int = 5,
              num_episodes_per_eval: int = 10,
              learning_rate: float = 0.001,
              train_context: core.EpisodesTrainContext = None,
              default_plots: bool = None):
        """Trains a new model using the gym environment passed during instantiation.

        Args:
            callbacks: list of callbacks called during training and evaluation
            num_iterations: number of times the training is repeated (with additional data)
            num_episodes_per_iteration: number of episodes played per training iteration
            max_steps_per_episode: maximum number of steps per episode
            num_epochs_per_iteration: number of times the data collected for the current iteration
                is used to retrain the current policy
            num_iterations_between_eval: number of training iterations before the current policy is evaluated.
                if 0 no evaluation is performed.
            num_episodes_per_eval: number of episodes played to estimate the average return and steps
            learning_rate: the learning rate used in the next iteration's policy training (0,1]
            train_context: training configuration to be used. if set overrides all other training context arguments.
            default_plots: if set adds a set of default callbacks (plot.State, plot.Rewards, plot.Loss,...).
                if None default callbacks are only added if the callbacks list is empty

        Returns:
            train_context: the training configuration containing the loss and sum of rewards encountered
                during training
        """
        if train_context is None:
            train_context = core.EpisodesTrainContext()
            train_context.num_iterations = num_iterations
            train_context.num_episodes_per_iteration = num_episodes_per_iteration
            train_context.max_steps_per_episode = max_steps_per_episode
            train_context.num_epochs_per_iteration = num_epochs_per_iteration
            train_context.num_iterations_between_eval = num_iterations_between_eval
            train_context.num_episodes_per_eval = num_episodes_per_eval
            train_context.learning_rate = learning_rate

        super().train(train_context=train_context, callbacks=callbacks, default_plots=default_plots)
        return train_context


class SacAgent(DqnAgent):
    """Agent based on the Soft-Actor-Critic algorithm (https://arxiv.org/abs/1812.05905)."""



import math
import fastrlapi.core as core


class Fast(core.AgentCallback):
    """Train for small number of episodes / steps in order to do a dry run of the algorithms or callbacks."""

    def __init__(self, num_iterations=None):
        self._num_iterations = num_iterations
        self._num_episodes_per_iteration = 3
        self._max_steps_per_episode = 50

    def on_play_begin(self, agent_context: core.AgentContext):
        agent_context.play.max_steps_per_episode = self._max_steps_per_episode
        if isinstance(agent_context, core.EpisodesTrainContext):
            agent_context.num_episodes_per_iteration = self._num_episodes_per_iteration

    def on_train_begin(self, agent_context: core.AgentContext):
        tc = agent_context.train
        if self._num_iterations is None:
            self._num_iterations = 10
        if isinstance(tc, core.EpisodesTrainContext):
            ec: core.EpisodesTrainContext = tc
            ec.num_episodes_per_iteration = self._num_episodes_per_iteration
            ec.num_epochs_per_iteration = 1
        if isinstance(tc, core.StepsTrainContext):
            dc: core.StepsTrainContext = tc
            if self._num_iterations is None:
                self._num_iterations = 5 * self._num_episodes_per_iteration * self._max_steps_per_episode
            dc.num_steps_buffer_preload = math.ceil(self._num_iterations / 10)
        tc.num_iterations = self._num_iterations
        tc.num_iterations_between_eval = math.ceil(tc.num_iterations/3)
        tc.num_episodes_per_eval = self._num_episodes_per_iteration
        tc.max_steps_per_episode = self._max_steps_per_episode


class _SingleEpisode(Fast):
    """Train / Play only for 1 episode (no evaluation in training, max. 10 steps)."""

    def __init__(self):
        super().__init__(num_iterations=1)
        self._num_episodes_per_iteration = 1
        self._max_steps_per_episode = 10

    def on_train_begin(self, agent_context: core.AgentContext):
        super().on_train_begin(agent_context)
        tc = agent_context.train
        if isinstance(tc, core.StepsTrainContext):
            tc.num_iterations = self._max_steps_per_episode
        tc.num_iterations_between_eval = 0
        tc.num_episodes_per_eval = 0


class _SingleIteration(Fast):
    """Train / play for a single iteration with 3 episodes, and evaluation of 2 episodes"""

    def __init__(self):
        super().__init__(num_iterations=1)



import unittest
import pytest
from typing import Dict, Type

import fastrlapi
import fastrlapi.core as core
import fastrlapi.env
import fastrlapi.backends.debug as debug
import fastrlapi.callbacks.log
import fastrlapi.backends.core as bcore


# noinspection PyTypeChecker
class BackendAgentTest(unittest.TestCase):
    env_name = fastrlapi.env._StepCountEnv.register_with_gym()

    def setUp(self):
        self.tc = core.EpisodesTrainContext()
        self.tc.num_episodes_per_iteration = 1
        self.tc.num_iterations = 1
        self.tc.num_episodes_per_eval = 2
        self.tc.max_steps_per_episode = 5
        self.pc = core.PlayContext(self.tc)

    class DebugAgent(debug.DebugAgent):
        def __init__(self):
            super().__init__(core.ModelConfig(gym_env_name=BackendAgentTest.env_name), action=1)

    def test_play_playcontext(self):
        agent = BackendAgentTest.DebugAgent()
        count = fastrlapi.callbacks.log._CallbackCounts()
        agent.play(play_context=self.pc, callbacks=[count])
        assert self.pc.play_done is True
        assert self.pc.episodes_done == self.pc.num_episodes
        assert self.pc.steps_done == self.pc.num_episodes * self.pc.max_steps_per_episode == 10
        assert self.pc.steps_done_in_episode == self.pc.max_steps_per_episode == 5

    def test_play_callbackinvariants(self):
        agent = BackendAgentTest.DebugAgent()
        agent.play(play_context=self.pc, callbacks=[debug.InvariantCallback()])

    def test_train_emptyArgs(self):
        agent = BackendAgentTest.DebugAgent()
        agent.train(train_context=self.tc, callbacks=[])
        assert self.tc.training_done
        assert self.tc.iterations_done_in_training == 1
        assert self.tc.episodes_done_in_iteration == 1
        assert self.tc.episodes_done_in_training == 1

    def test_train_missingArgs(self):
        agent = BackendAgentTest.DebugAgent()
        with self.assertRaises(AssertionError):
            agent.train(train_context=None, callbacks=[])
        with self.assertRaises(AssertionError):
            agent.train(train_context=self.tc, callbacks=None)

    def test_train_traincontext(self):
        agent = BackendAgentTest.DebugAgent()
        count = fastrlapi.callbacks.log._CallbackCounts()
        tc = self.tc
        tc.num_iterations = 3
        tc.num_episodes_per_iteration = 2
        tc.max_steps_per_episode = 10
        tc.num_episodes_per_eval = 5
        tc.num_iterations_between_eval = 2
        agent.train(train_context=tc, callbacks=[count])
        assert count.train_begin_count == count.train_end_count == 1
        assert count.train_iteration_begin_count == count.train_iteration_end_count == 3
        assert tc.episodes_done_in_training == 6
        assert tc.episodes_done_in_iteration == 2
        assert tc.episodes_done_in_training in tc.loss
        assert tc.steps_done_in_training == tc.num_iterations * tc.num_episodes_per_iteration * \
               tc.max_steps_per_episode
        assert tc.steps_done_in_iteration == tc.num_episodes_per_iteration * tc.max_steps_per_episode

        assert 0 in tc.eval_rewards
        assert 0 in tc.eval_steps
        assert tc.num_iterations_between_eval * tc.num_episodes_per_iteration in tc.eval_rewards
        assert tc.num_iterations_between_eval * tc.num_episodes_per_iteration in tc.eval_steps
        assert tc.episodes_done_in_training in tc.eval_rewards
        assert tc.episodes_done_in_training in tc.eval_steps

        assert count.gym_init_begin_count == count.gym_init_end_count > 0
        assert count.gym_reset_begin_count == count.gym_reset_end_count > 0
        assert count.gym_step_begin_count == count.gym_step_end_count > 0

    def test_save_directory_does_not_exist(self):
        tempdir = bcore._get_temp_path()
        agent = BackendAgentTest.DebugAgent()
        with pytest.raises(Exception):
            agent.save(tempdir, [])


    def test_save_directory_exists(self):
        tempdir = bcore._get_temp_path()
        fastrlapi.backends.core._mkdir(tempdir)
        agent = BackendAgentTest.DebugAgent()
        tc = core.EpisodesTrainContext()
        tc.num_iterations=1
        agent.train(callbacks=[], train_context=tc)
        agent.save(tempdir, [])
        fastrlapi.backends.core._rmpath(tempdir)


class BackendAgentFactoryTest(unittest.TestCase):
    class DebugAgentFactory(bcore.BackendAgentFactory):
        class DebugAgent(debug.DebugAgent):
            def __init__(self, model_config: core.ModelConfig):
                super().__init__(model_config=model_config, action=1)

        backend_name = "debug"

        def get_algorithms(self) -> Dict[Type[fastrlapi.agents.EasyAgent], Type[bcore._BackendAgent]]:
            return {fastrlapi.agents.DqnAgent: BackendAgentFactoryTest.DebugAgentFactory.DebugAgent}

    def test_get_algorithms(self):
        f = BackendAgentFactoryTest.DebugAgentFactory()
        a = f.get_algorithms()
        assert fastrlapi.agents.DqnAgent in a

    def test_create_agent(self):
        f = BackendAgentFactoryTest.DebugAgentFactory()
        mc = core.ModelConfig(gym_env_name="CartPole-v0")
        a = f.create_agent(easyagent_type=fastrlapi.agents.DqnAgent, model_config=mc)
        assert a is not None

    def test_create_agent_not_implemented(self):
        f = BackendAgentFactoryTest.DebugAgentFactory()
        mc = core.ModelConfig(gym_env_name="CartPole-v0")
        a = f.create_agent(easyagent_type=fastrlapi.agents.ReinforceAgent, model_config=mc)
        assert a is None

    def test_get_backend(self):
        b = fastrlapi.agents._get_backend('tfagents')
        assert b

    def test_get_backend_invalid_name(self):
        assert fastrlapi.agents._get_backend('noname') is None



"""This module contains the backend implementation for tensorforce

    see https://github.com/tensorforce/tensorforce
"""
import datetime
import math
import os
import tempfile
from abc import ABCMeta
from typing import List, Dict, Optional, Type

import gym
from tensorforce.agents import Agent
from tensorforce.environments import Environment
from tensorforce.execution import Runner

import fastrlapi.backends.core


class TforceAgent(fastrlapi.backends.core.BackendAgent, metaclass=ABCMeta):
    """ Base class for agents based on the tensorforce implementation."""

    def __init__(self, model_config: fastrlapi.core.ModelConfig):
        """
        Args:
            model_config: the model configuration including the name of the target gym environment
                as well as the neural network architecture.
        """
        super().__init__(model_config=model_config,
                         backend_name=TensorforceAgentFactory.backend_name,
                         tf_eager_execution=False)
        self._agent: Optional[Agent] = None
        self._play_env: Optional[Environment] = None

    def _create_env(self) -> Environment:
        """Creates a tensorforce Environment encapsulating the underlying gym environment given in self.model_config"""
        self.log_api('Environment.create', f'(environment="gym", level={self.model_config.original_env_name})')
        result = Environment.create(environment='gym', level=self.model_config.gym_env_name)
        return result

    def _create_network_specification(self):
        """Creates a tensorforce network specification based on the layer specification given in self.model_config"""
        result: List[Dict] = []
        layer_sizes = self.model_config.fc_layers
        for layer_size in layer_sizes:
            result.append(dict(type='dense', size=layer_size, activation='relu'))
        return result

    def _get_temp_path(self):
        result = os.path.join(tempfile.gettempdir(), tempfile.gettempprefix())
        n = datetime.datetime.now()
        result = result + f'-{n.year % 100:2}{n.month:02}{n.day:02}-{n.hour:02}{n.minute:02}{n.second:02}-' + \
                 f'{n.microsecond:06}'
        return result

    def _train_with_runner(self, train_env: Environment, train_context: fastrlapi.core.TrainContext):
        """Trains the self._agent using a tensorforce runner.

        Args:
            train_env: the tensorforce environment to use for the training
            train_context: context containing the training parameters
        """
        assert train_context
        assert train_env
        assert self._agent

        def step_callback(tforce_runner: Runner) -> bool:
            """
                Returns:
                    true if the training should continue, false to end the training
            """
            result = not train_context.training_done
            if result:
                is_iteration_end = False
                if isinstance(train_context, fastrlapi.core.EpisodesTrainContext):
                    ec: fastrlapi.core.EpisodesTrainContext = train_context
                    current_episode = tforce_runner.episodes - 1
                    current_step = tforce_runner.episode_timestep
                    is_iteration_end = current_episode > 0 and \
                                       current_step == 1 and \
                                       current_episode % ec.num_episodes_per_iteration == 0
                elif isinstance(train_context, fastrlapi.core.StepsTrainContext):
                    sc: fastrlapi.core.StepsTrainContext = train_context
                    current_step = tforce_runner.timesteps - 1
                    is_iteration_end = current_step > 0 and \
                                       current_step % sc.num_steps_per_iteration == 0
                else:
                    raise AssertionError("Unexpected TrainContext detected.")
                if is_iteration_end:
                    self.on_train_iteration_end(loss=math.nan, actor_loss=math.nan, critic_loss=math.nan)
                    result = not train_context.training_done
                    if result:
                        self.on_train_iteration_begin()
            return result

        def eval_callback(tforce_runner: Runner) -> bool:
            """
                This call back should never be called. Raises an exception.

                Policy evaluation is performed by fastrlapi directly.
            """
            raise AssertionError("unexpected eval_callback call.")

        # Initialize the runner
        self.log_api('Runner.create', "(agent=..., environment=...)")
        runner = Runner(agent=self._agent, environment=train_env)

        # Start the runner
        self.on_train_iteration_begin()
        self.log_api('runner.run',
                     f'(num_episodes=None, use_tqdm=False, callback=..., callback_timestep_frequency=1, ' +
                     'evaluation_callback=..., evaluation_frequency=None, evaluation=False, ' +
                     'num_evaluation_iterations=0)')
        runner.run(num_episodes=None,
                   use_tqdm=False,
                   callback=step_callback,
                   callback_timestep_frequency=1,
                   evaluation_callback=eval_callback,
                   evaluation_frequency=None,
                   evaluation=False,
                   num_evaluation_iterations=0
                   )
        assert train_context.training_done
        "Unexpected runner termination."
        runner.close()

    def load_implementation(self, directory: str):
        """Loads a previously trained and saved actor policy from directory.

        The loaded policy may afterwards be used by calling play().

        Args:
            directory: the directory containing the trained policy.
        """
        assert directory
        assert os.path.isdir(directory)
        self._agent = Agent.load(directory)

    def play_implementation(self, play_context: fastrlapi.core.PlayContext):
        """Agent specific implementation of playing a single episode with the current policy.

            Args:
                play_context: play configuration to be used
        """
        assert play_context, "play_context not set."
        assert self._agent, "agent not set. call train first."

        if self._play_env is None:
            self._play_env = self._create_env()
        while True:
            # noinspection PyUnresolvedReferences
            gym_env: gym.Env = self._play_env.environment
            self.on_play_episode_begin(env=gym_env)
            state = self._play_env.reset()
            done = False
            while not done:
                action = self._agent.act(state, evaluation=True)
                state, terminal, reward = self._play_env.execute(actions=action)
                if isinstance(terminal, bool):
                    done = terminal
                else:
                    done = terminal > 0
            self.on_play_episode_end()
            if play_context.play_done:
                break
        return

    def save_implementation(self, directory: str):
        """Agent speecific implementation of saving the weights for the actor policy.

        Save must only guarantee to persist the weights of the actor policy.
        The implementation may write multiple files with fixed filenames.

        Args:
             directory: the (existing) directory to save the policy weights to.
        """
        assert directory
        assert os.path.isdir(directory)
        self._agent.save(directory)


class TforceDqnAgent(TforceAgent):
    """ Agent based on the DQN algorithm using the tensorforce implementation."""

    def __init__(self, model_config: fastrlapi.core.ModelConfig,
                 enable_dueling_dqn: bool = False):
        """
        Args:
            model_config: the model configuration including the name of the target gym environment
                as well as the neural network architecture.
        """
        super().__init__(model_config=model_config)
        self._agent_type : str = 'dueling_dqn' if enable_dueling_dqn else 'dqn'

    def train_implementation(self, train_context: fastrlapi.core.StepsTrainContext):
        """Tensorforce Dqn Implementation of the train loop.

            The implementation follows https://github.com/tensorforce/tensorforce/blob/master/examples/quickstart.py
        """
        tc = train_context
        train_env = self._create_env()
        network = self._create_network_specification()
        memory = tc.max_steps_in_buffer + tc.num_steps_sampled_from_buffer
        exploration=0.1
        self.log_api('Agent.create',
                     f'(agent="{self._agent_type}", ' +
                     f'network={network}, ' +
                     f'memory={memory}, ' +
                     f'start_updating={tc.num_steps_buffer_preload},'
                     f'learning_rate={tc.learning_rate}, ' +
                     f'batch_size={tc.num_steps_sampled_from_buffer}, ' +
                     f'update_frequeny={tc.num_steps_per_iteration}, ' +
                     f'discount={tc.reward_discount_gamma}, ' +
                     f'exploration={exploration}')
        self._agent = Agent.create(
            agent=self._agent_type,
            environment=train_env,
            network=network,
            memory=memory,
            start_updating=tc.num_steps_buffer_preload,
            learning_rate=tc.learning_rate,
            batch_size=tc.num_steps_sampled_from_buffer,
            update_frequency=tc.num_steps_per_iteration,
            discount=tc.reward_discount_gamma,
            exploration=exploration,
            preprocessing = None,
            variable_noise = 0.0,
            l2_regularization = 0.0,
            entropy_regularization = 0.1
        )
        self._train_with_runner(train_env, tc)

class TforceDuelingDqnAgent(TforceDqnAgent):
    """ Agent based on the DQN algorithm using the tensorforce implementation."""

    def __init__(self, model_config: fastrlapi.core.ModelConfig):
        """
        Args:
            model_config: the model configuration including the name of the target gym environment
                as well as the neural network architecture.
        """
        super().__init__(model_config=model_config, enable_dueling_dqn=True)


class TforcePpoAgent(TforceAgent):
    """ Agent based on the PPO algorithm using the tensorforce implementation."""

    def train_implementation(self, train_context: fastrlapi.core.PpoTrainContext):
        """Tensorforce Ppo Implementation of the train loop.

            The implementation follows https://github.com/tensorforce/tensorforce/blob/master/examples/quickstart.py
        """
        tc = train_context
        train_env = self._create_env()
        network = self._create_network_specification()

        self.log_api('Agent.create', f'(agent="ppo", environment=..., ' +
                     f'network={network}' +
                     f'learning_rate={tc.learning_rate}, ' +
                     f'batch_size={tc.num_episodes_per_iteration}, ' +
                     f'optimization_steps={tc.num_epochs_per_iteration}, ' +
                     f'discount={tc.reward_discount_gamma})')
        self._agent = Agent.create(
            agent='ppo',
            environment=train_env,
            network=network,
            learning_rate=tc.learning_rate,
            batch_size=tc.num_episodes_per_iteration,
            optimization_steps=tc.num_epochs_per_iteration,
            discount=tc.reward_discount_gamma,
        )
        self._train_with_runner(train_env, tc)


class TforceRandomAgent(TforceAgent):
    """ Random agent using the tensorforce implementation."""

    def train_implementation(self, train_context: fastrlapi.core.TrainContext):
        assert isinstance(train_context, fastrlapi.core.TrainContext)
        train_env = self._create_env()
        self.log_api('Agent.create', f'(agent="random", environment=...)')
        self._agent = Agent.create(agent='random', environment=train_env)
        if not self._agent.model.is_initialized:
            self._agent.initialize()

        while not train_context.training_done:
            self.on_train_iteration_begin()
            state = train_env.reset()
            done = False
            while not done:
                action = self._agent.act(state, evaluation=True)
                state, terminal, reward = train_env.execute(actions=action)
                if isinstance(terminal, bool):
                    done = terminal
                else:
                    done = terminal > 0
            self.on_train_iteration_end(math.nan)


class TforceReinforceAgent(TforceAgent):
    """ Agent based on the REINFORCE algorithm using the tensorforce implementation."""

    def train_implementation(self, train_context: fastrlapi.core.EpisodesTrainContext):
        """Tensorforce REINFORCE Implementation of the train loop.

            The implementation follows https://github.com/tensorforce/tensorforce/blob/master/examples/quickstart.py
        """
        assert isinstance(train_context, fastrlapi.core.EpisodesTrainContext)
        tc = train_context
        train_env = self._create_env()
        network = self._create_network_specification()

        self.log_api('Agent.create', f'(agent="vpg", environment=..., ' +
                     f'network={network}' +
                     f'learning_rate={tc.learning_rate}, ' +
                     f'batch_size={tc.num_episodes_per_iteration}, ' +
                     f'discount={tc.reward_discount_gamma})')
        self._agent = Agent.create(
            agent='vpg',
            environment=train_env,
            network=network,
            learning_rate=tc.learning_rate,
            batch_size=tc.num_episodes_per_iteration,
            discount=tc.reward_discount_gamma,
        )
        self._train_with_runner(train_env, tc)


class TensorforceAgentFactory(fastrlapi.backends.core.BackendAgentFactory):
    """Backend for Tensorforce.

        Serves as a factory to create algorithm specific wrappers for the tensorforce implementations.
    """

    backend_name: str = 'tensorforce'

    def get_algorithms(self) -> Dict[Type, Type[fastrlapi.backends.core.BackendAgent]]:
        """Yields a mapping of EasyAgent types to the implementations provided by this backend."""
        return {
            fastrlapi.agents.DqnAgent: TforceDqnAgent,
            fastrlapi.agents.DuelingDqnAgent: TforceDuelingDqnAgent,
            fastrlapi.agents.PpoAgent: TforcePpoAgent,
            fastrlapi.agents.RandomAgent: TforceRandomAgent,
            fastrlapi.agents.ReinforceAgent: TforceReinforceAgent}



"""Contains the implementation of the gym environment monitor used to intercept all calls from
   a BackendAgent to the gym environment. The monitor is the basis for ApiCallbacks as well as
   the gym environment statistics like the number of episodes played or the number of steps
   taken during training.
"""

from typing import Dict, Optional
import threading

import gym.core
import gym.envs
import fastrlapi.backends


class _MonitorTotalCounts(object):
    """Keeps usage counts on a monitored gym environment.

    Attributes:
        gym_env_name: the id of the monitored gym environment
    """

    def __init__(self, gym_env_name):
        assert isinstance(gym_env_name, str), "gym_env_name not a str"
        assert gym_env_name, "gym_env_name empty"

        self._original_env_name = gym_env_name
        self.gym_env_name = _MonitorEnv._NAME_PREFIX + self._original_env_name
        self._instances_created: int = 0
        self._episodes_done: int = 0
        self._steps_done: int = 0

    def __str__(self):
        return f'[{self._original_env_name}] #instances={self.instances_created} ' + \
               f'#episodes={self.episodes_done} #steps={self.steps_done}'

    @property
    def instances_created(self):
        """the total number of instances created"""
        with _MonitorEnv._lock:
            return self._instances_created

    def instances_created_inc(self):
        """increments the total number of instances created"""
        with _MonitorEnv._lock:
            self._instances_created += 1
            return self._instances_created

    @property
    def episodes_done(self):
        """the total number of episodes completed (over all instances).

            A new episodes starts with a new instance or a call to reset.
            If step() was not called after the last reset, than the episode count is not incremented.
            Thus calling reset() multiple times, or calling reset right after instantiation increments
            the episode count at most by 1.
        """
        with _MonitorEnv._lock:
            return self._episodes_done

    def episodes_done_inc(self):
        """increments the total number of new episodes (over all instances).
        """
        with _MonitorEnv._lock:
            self._episodes_done += 1
            return self._episodes_done

    @property
    def steps_done(self):
        """the number of step completed (over all instances)"""
        with _MonitorEnv._lock:
            return self._steps_done

    def steps_done_inc(self):
        """increments the total number of step() calls (over all instances)"""
        with _MonitorEnv._lock:
            self._steps_done += 1
            return self._steps_done


class _MonitorEnv(gym.Wrapper):
    """Intercepts all calls between a BackendAgent and its gym environment.

    Attributes:
        gym_env_name: the name of the monitored gym env
        instance: unique id among all monitor instances for the given gym_env
        total: statistics over all monitor instances for the given gym_env
        episodes_done: the number of episodes completed in this instance (episode reached done or was resetted)
        is_episode_done: true if the current episode is done (and a new episode isn't started yet with reset())
        steps_done_in_episode: the number of steps completed in the current episode
        steps_done_in_instance: the total number of steps in this instance
        max_steps_per_episode: the max number of steps in an episode (None if unlimited)
    """

    _NAME_PREFIX = "_MonitorEnv_"
    _KWARG_GYM_ENV_NAME = "fastrlapi_gym_env_name"
    _monitor_total_counts: Dict[str, _MonitorTotalCounts] = dict()
    _lock = threading.Lock()
    _backend_agent = None

    @staticmethod
    def _register_backend_agent(backend_agent):
        """Registers backend_agent als callback target for all new monitored gym instances.

            All newly created MonitorEnv will call backend_agent upon step / reset calls.
        """
        with _MonitorEnv._lock:
            if _MonitorEnv._backend_agent is None or backend_agent is None:
                _MonitorEnv._backend_agent = backend_agent
            else:
                assert _MonitorEnv._backend_agent is backend_agent, "another backend agent is already registered"
                _MonitorEnv._backend_agent = backend_agent

    def __init__(self, **kwargs):
        assert _MonitorEnv._KWARG_GYM_ENV_NAME in kwargs, f'{_MonitorEnv._KWARG_GYM_ENV_NAME} missing from kwargs'

        self.gym_env_name = kwargs[_MonitorEnv._KWARG_GYM_ENV_NAME]
        with _MonitorEnv._lock:
            self.total = _MonitorEnv._monitor_total_counts[self.gym_env_name]
        self.steps_done_in_instance: int = 0
        self.episodes_done: int = 0
        self.steps_done_in_episode: int = 0
        self.max_steps_per_episode: Optional[int] = None
        self.is_episode_done: bool = False
        self.instance_id = self.total.instances_created
        self._backend_agent: Optional[fastrlapi.backends.core._BackendAgent] = _MonitorEnv._backend_agent

        if self._backend_agent:
            self._backend_agent._on_gym_init_begin()
        gym_env = gym.make(self.gym_env_name)
        super().__init__(gym_env)
        self.total.instances_created_inc()
        if self._backend_agent:
            self._backend_agent._on_gym_init_end(self)

    def reset(self, **kwargs):
        """performs a reset on the wrapped environment."""
        if self._backend_agent:
            self._backend_agent._on_gym_reset_begin(self, **kwargs)

        result = self.env.reset(**kwargs)
        if self.steps_done_in_episode > 0 and not self.is_episode_done:
            self.episodes_done += 1
            self.total.episodes_done_inc()
        self.is_episode_done = False
        self.steps_done_in_episode = 0

        if self._backend_agent:
            self._backend_agent._on_gym_reset_end(self, result, **kwargs)
        return result

    def step(self, action):
        """performs a step on the wrapped environment.

        Hint:
        o episode is terminated if steps_done_in_episode >= max_steps_per_episode

        Returns:
            (state, reward, done, info)
        """
        if self._backend_agent:
            self._backend_agent._on_gym_step_begin(self, action)

        result = self.env.step(action)
        (state, reward, done, info) = result
        self.steps_done_in_episode += 1
        self.steps_done_in_instance += 1
        self.total.steps_done_inc()
        if self.max_steps_per_episode and self.steps_done_in_episode >= self.max_steps_per_episode:
            done = True
            result = (state, reward, done, info)
        if not self.is_episode_done and done:
            self.is_episode_done = True
            self.episodes_done += 1
            self.total.episodes_done_inc()

        if self._backend_agent:
            self._backend_agent._on_gym_step_end(self, action, result)
        return result

    def __str__(self):
        return f'[{self.gym_env_name}#{self.instance_id:}] {self.episodes_done:3}#{self.steps_done_in_episode:<3} '


def _get(env: gym.Env) -> _MonitorEnv:
    """extracts from env the underlying _MonitorEnv, returns None of not successful"""
    assert env, "env not set"
    result = None
    if isinstance(env, _MonitorEnv):
        result = env
    else:
        if isinstance(env, gym.core.Wrapper):
            result = _get(env.env)
    return result


def _register_gym_monitor(gym_env_name: str) -> _MonitorTotalCounts:
    """Registers the _MonitorEnv wrapper for the 'gym_env_name' environment.

    The wrapper is registered as '_MonitorEnv-<env_name>'.
    If the same name is registered more than once, the second and all following calls are noOps.

    Args:
        gym_env_name: the gym id of the environment to be monitored

    Retuns:
        the count monitor of gym_env_name
    """
    assert gym_env_name, "gym_env_name must be a non-empty string"
    assert type(gym_env_name) is str, "gym_env_name is not a str"

    result: _MonitorTotalCounts
    with _MonitorEnv._lock:
        if gym_env_name not in _MonitorEnv._monitor_total_counts:
            result = _MonitorTotalCounts(gym_env_name)

            gym_spec: Optional[gym.envs.registration.EnvSpec] = None
            try:
                gym_spec = gym.envs.registration.spec(gym_env_name)
            except Exception:
                pass
            assert gym_spec, f'no registration found for gym environment {gym_env_name}'

            # max_episode_steps = None results in a tfagents gym_suit.load crash
            gym.envs.registration.register(id=result.gym_env_name,
                                           entry_point=_MonitorEnv,
                                           max_episode_steps=gym_spec.max_episode_steps,
                                           kwargs={_MonitorEnv._KWARG_GYM_ENV_NAME: gym_env_name})
            _MonitorEnv._monitor_total_counts[gym_env_name] = result
        else:
            result = _MonitorEnv._monitor_total_counts[gym_env_name]
    return result



import unittest

import pytest

import fastrlapi.backends.core
import fastrlapi.backends.debug
from fastrlapi.agents import PpoAgent
from fastrlapi.callbacks.duration import Fast
from fastrlapi.core import ModelConfig, AgentCallback, AgentContext, PpoTrainContext, PlayContext

_stepcount_name = fastrlapi.env._StepCountEnv.register_with_gym()


class AgentContextTest(unittest.TestCase):
    class PlayCallback(AgentCallback):
        def __init__(self):
            self.play_called = False
            self.train_called = False

        def on_play_begin(self, ac: AgentContext):
            assert ac.is_play
            assert not ac.is_eval
            assert not ac.is_train
            self.play_called = True

        def on_train_begin(self, agent_context: AgentContext):
            self.train_called = True

    class TrainCallback(AgentCallback):

        def __init__(self):
            self.play_called = False
            self.train_called = False

        def on_play_begin(self, agent_context: AgentContext):
            assert not agent_context.is_play
            assert agent_context.is_eval
            assert not agent_context.is_train
            self.play_called = True

        def on_train_begin(self, agent_context: AgentContext):
            assert not agent_context.is_play
            assert not agent_context.is_eval
            assert agent_context.is_train
            self.train_called = True

    def test_agentcontext_train(self):
        b = fastrlapi.backends.debug.DebugAgentFactory()
        a = b.create_agent(PpoAgent, ModelConfig(_stepcount_name))
        c = AgentContextTest.TrainCallback()
        a.train(callbacks=[Fast(), c], train_context=PpoTrainContext())
        assert c.train_called
        assert c.play_called

    def test_agentcontext_play(self):
        b = fastrlapi.backends.debug.DebugAgentFactory()
        a = b.create_agent(PpoAgent, ModelConfig(_stepcount_name))
        c = AgentContextTest.PlayCallback()
        pc = PlayContext()
        pc.num_episodes = 10
        pc.max_steps_per_episode = 10
        a.play(callbacks=[Fast(), c], play_context=pc)
        assert not c.train_called
        assert c.play_called


class ModelConfigTest(unittest.TestCase):

    def test_create(self):
        assert ModelConfig(gym_env_name=_stepcount_name) is not None
        assert ModelConfig(gym_env_name=_stepcount_name, fc_layers=(10, 20)) is not None

    def test_create_envNotRegistered_exception(self):
        with pytest.raises(AssertionError):
            ModelConfig(gym_env_name="MyEnv-v0")

    def test_create_envNotnameNotSet_exception(self):
        with pytest.raises(AssertionError):
            # noinspection PyTypeChecker
            ModelConfig(gym_env_name=None)

    def test_create_fclayersEmpty_exception(self):
        with pytest.raises(AssertionError):
            ModelConfig(gym_env_name=_stepcount_name, fc_layers=())

    def test_create_fclayersSimpleInt(self):
        assert ModelConfig(gym_env_name=_stepcount_name, fc_layers=10) is not None

    def test_create_fclayersNegativeValue(self):
        with pytest.raises(AssertionError):
            ModelConfig(gym_env_name=_stepcount_name, fc_layers=-10)

    def test_to_from_dict(self):
        mc = ModelConfig(gym_env_name=_stepcount_name, fc_layers=10,seed=123)
        mc_dict = mc._to_dict()
        mc2 = ModelConfig._from_dict(mc_dict)
        assert mc2.fc_layers == mc.fc_layers
        assert mc2.original_env_name == mc.original_env_name
        assert mc2.seed == mc.seed

if __name__ == '__main__':
    unittest.main()



"""This module contains the backend for a constant action agent implementation.

    All agents created by the debug backend execute a "no operations" train loop,
    without an underlying model / neural network.
"""

from typing import Tuple, Dict, Type
from fastrlapi import core
from fastrlapi.backends import core as bcore
import fastrlapi.agents
import gym


class DebugAgentFactory(bcore.BackendAgentFactory):
    backend_name: str = 'debug'

    def get_algorithms(self) -> Dict[Type, Type[bcore._BackendAgent]]:
        """Yields a mapping of EasyAgent types to the implementations provided by this backend."""
        return {fastrlapi.agents.DqnAgent: DebugAgent,
                fastrlapi.agents.PpoAgent: DebugAgent,
                fastrlapi.agents.RandomAgent: DebugAgent,
                fastrlapi.agents.ReinforceAgent: DebugAgent}


class DebugAgent(bcore._BackendAgent):

    def __init__(self, model_config: core.ModelConfig, action=None):
        """Simple constant action agent.

        Args:
            model_config: containing the gym_env_name to "train" on
            action: the action to take in all steps or None. If None no steps are taken.
        """
        super().__init__(model_config, backend_name=DebugAgentFactory.backend_name, tf_eager_execution=True)
        self.action = action

    def play_implementation(self, play_context: core.PlayContext):
        env = gym.make(self.model_config.gym_env_name)
        while True:
            self.on_play_episode_begin(env=env)
            env.reset()
            if self.action is not None:
                done = False
                while not done:
                    (observation, reward, done, info) = env.step(self.action)
            self.on_play_episode_end()
            if play_context.play_done:
                break

    def train_implementation(self, train_context: core.TrainContext):
        assert isinstance(train_context, core.EpisodesTrainContext)
        tc: core.EpisodesTrainContext = train_context

        env = gym.make(self.model_config.gym_env_name)
        for i in range(tc.num_iterations):
            self.on_train_iteration_begin()
            for e in range(tc.num_episodes_per_iteration):
                env.reset()
                if self.action is not None:
                    done = False
                    while not done:
                        (observation, reward, done, info) = env.step(self.action)
            self.on_train_iteration_end(0)
            if tc.training_done:
                break

    def load_implementation(self, directory: str):
        pass

    def save_implementation(self, directory: str):
        pass


class InvariantCallback(core.AgentCallback):
    """Validates the callback invariants"""

    def on_play_begin(self, agent_context: core.AgentContext):
        assert agent_context.play.gym_env is None

    def on_play_end(self, agent_context: core.AgentContext):
        assert agent_context.play.gym_env is not None

    def on_play_episode_begin(self, agent_context: core.AgentContext):
        assert agent_context.play.gym_env is not None

    def on_play_episode_end(self, agent_context: core.AgentContext):
        assert agent_context.play.gym_env is not None

    def on_play_step_begin(self, agent_context: core.AgentContext, action):
        assert agent_context.play.gym_env is not None

    def on_play_step_end(self, agent_context: core.AgentContext, action, step_result: Tuple):
        assert agent_context.play.gym_env is not None



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="fastrlapi",
    version="0.0.1",
    description="reinforcement learning high-level API.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="",
    author="Rohit Gangupantulu",
    author_email="rgangupantulu@deloitte.com",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["fastrlapi", "fastrlapi.callbacks", "fastrlapi.backends"],
    install_requires=[
        "tensorflow==2.0.1",
        "tensorflow-probability==0.8.0",
        "tf-agents==0.3.0",
        "tensorforce==0.5.3",
        "gym==0.15.4",
        "imageio==2.6.1",
        "imageio-ffmpeg==0.3.0",
        "matplotlib==3.1.2"
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "drlapi=drlapi.__main__:main",
        ]
    },
)



import unittest
import gym
from fastrlapi.env import _is_registered_with_gym, register_with_gym, _LineWorldEnv

_line_world_name = _LineWorldEnv.register_with_gym()


class TestLineWorld(unittest.TestCase):
    def test_is_registered_with_gym(self):
        assert _is_registered_with_gym('LineWorld-v0')

    def test_make(self):
        lw = gym.make(_line_world_name)
        assert isinstance(lw.action_space, gym.spaces.Discrete)
        assert isinstance(lw.observation_space, gym.spaces.Box)

    def test_step_right_left(self):
        lw = gym.make(_line_world_name)
        obs = lw.reset()
        assert obs[0] == 3
        obs, reward, done, info = lw.step(1)
        assert obs[0] == 4
        assert reward == -1
        assert done == False
        obs, reward, done, info = lw.step(0)
        assert obs[0] == 3
        assert reward == 4
        assert done == False

    def test_step_left(self):
        lw = gym.make(_line_world_name)
        lw.reset()
        count = 0
        while True:
            o, r, d, i = lw.step(0)
            count += 1
            if d:
                break
        assert count == 3

    def test_step_right(self):
        lw = gym.make(_line_world_name)
        lw.reset()
        count = 0
        while True:
            o, r, d, i = lw.step(1)
            count += 1
            if d:
                break
        assert count == 37


class TestEnv(unittest.TestCase):
    def test_is_registered_with_gym(self):
        assert _is_registered_with_gym('CartPole-v0')

    def test_is_registered_with_gym_missingRegistration(self):
        assert not _is_registered_with_gym('MyEnv-v0')


class TestShimEnv(unittest.TestCase):

    def test_register_once_no_args(self):
        register_with_gym("test_env-v0", _Env1)
        env1 = gym.make("test_env-v0")
        assert isinstance(env1.unwrapped, _Env1)

    def test_register_once_with_args(self):
        register_with_gym("test_env-v0", _Env1, a1=1, a2=2)
        env1 = gym.make("test_env-v0")
        assert env1.unwrapped.a1 == 1

    def test_register_twice(self):
        register_with_gym("test_env-v0", _Env1)
        register_with_gym("test_env-v0", _Env2)
        env2 = gym.make("test_env-v0")
        assert isinstance(env2.unwrapped, _Env2)

    def test_register_twice_with_args(self):
        register_with_gym("test_env-v0", _Env1, a1=1, a2=2)
        register_with_gym("test_env-v0", _Env1, a1=3, a2=4)
        env1 = gym.make("test_env-v0")
        assert env1.unwrapped.a1 == 3

    def test_gym(self):
        gym.envs.registration.register(id="test_env-v1", entry_point=_Env1)
        gym.make("test_env-v1")


class _Env1(gym.Env):
    def __init__(self, a1=0, **kwargs):
        self.a1 = a1
        self.kwargs = kwargs
        pass

    def render(self, mode='human'):
        pass

    def reset(self):
        pass

    def step(self, action):
        pass


class _Env2(gym.Env):
    def __init__(self):
        pass

    def render(self, mode='human'):
        pass

    def reset(self):
        pass

    def step(self, action):
        pass


if __name__ == '__main__':
    unittest.main()



import unittest

from fastrlapi import core, env, agents
from fastrlapi.backends import tfagents
from fastrlapi.backends import core as bcore
from fastrlapi.callbacks import duration, log

agents.seed = 0
_lineworld_name = env._LineWorldEnv.register_with_gym()
_stepcount_name = env._StepCountEnv.register_with_gym()
_mountaincar_continuous_name = 'MountainCarContinuous-v0'


class TfDqnAgentTest(unittest.TestCase):

    def setUp(self):
        self.env_name = env._StepCountEnv.register_with_gym()

    def test_train(self):
        model_config = core.ModelConfig("CartPole-v0")
        tc = core.StepsTrainContext()
        dqn_agent = tfagents.TfDqnAgent(model_config=model_config)
        dqn_agent.train(train_context=tc, callbacks=[duration.Fast(), log.Iteration()])


class TfPpoAgentTest(unittest.TestCase):

    def setUp(self):
        self.env_name = env._StepCountEnv.register_with_gym()

    def test_train(self):
        model_config = core.ModelConfig(_lineworld_name)
        tc = core.PpoTrainContext()
        ppo_agent = tfagents.TfPpoAgent(model_config=model_config)
        ppo_agent.train(train_context=tc, callbacks=[duration.Fast(), log.Iteration()])

    def test_save_load(self):
        model_config = core.ModelConfig(_lineworld_name)
        tc = core.PpoTrainContext()
        ppo_agent = tfagents.TfPpoAgent(model_config=model_config)
        ppo_agent.train(train_context=tc, callbacks=[duration._SingleIteration(), log.Iteration()])
        tempdir = bcore._get_temp_path()
        bcore._mkdir(tempdir)
        ppo_agent.save(tempdir, [])
        ppo_agent = tfagents.TfPpoAgent(model_config=model_config)
        ppo_agent.load(tempdir, [])
        pc = core.PlayContext()
        pc.max_steps_per_episode = 10
        pc.num_episodes = 1
        ppo_agent.play(play_context=pc, callbacks=[])
        bcore._rmpath(tempdir)


class TfRandomAgentTest(unittest.TestCase):

    def test_train(self):
        model_config = core.ModelConfig(_lineworld_name)
        tc = core.TrainContext()
        random_agent = tfagents.TfRandomAgent(model_config=model_config)
        random_agent.train(train_context=tc, callbacks=[duration.Fast(), log.Iteration()])
        assert tc.episodes_done_in_iteration == 1

    def test_play(self):
        model_config = core.ModelConfig(_lineworld_name)
        random_agent = tfagents.TfRandomAgent(model_config=model_config)
        pc = core.PlayContext()
        pc.max_steps_per_episode = 10
        pc.num_episodes = 1
        random_agent.play(play_context=pc, callbacks=[])
        assert pc.num_episodes == 1

    def test_save_load(self):
        model_config = core.ModelConfig(_lineworld_name)
        random_agent = tfagents.TfRandomAgent(model_config=model_config)
        tempdir = bcore._get_temp_path()
        bcore._mkdir(tempdir)
        random_agent.save(directory=tempdir, callbacks=[])
        random_agent.load(directory=tempdir, callbacks=[])
        bcore._rmpath(tempdir)


class TfReinforceAgentTest(unittest.TestCase):

    def test_train(self):
        model_config = core.ModelConfig(_lineworld_name)
        tc = core.EpisodesTrainContext()
        reinforce_agent = tfagents.TfReinforceAgent(model_config=model_config)
        reinforce_agent.train(train_context=tc, callbacks=[duration.Fast(), log.Iteration()])
        assert tc.episodes_done_in_iteration == tc.num_episodes_per_iteration > 0
        assert tc.iterations_done_in_training == tc.num_iterations > 0


class TfSacAgentTest(unittest.TestCase):

    def setUp(self):
        self.env_name = env._StepCountEnv.register_with_gym()

    def test_train(self):
        model_config = core.ModelConfig(_mountaincar_continuous_name)
        tc = core.StepsTrainContext()
        dqn_agent = tfagents.TfSacAgent(model_config=model_config)
        dqn_agent.train(train_context=tc, callbacks=[duration.Fast(), log.Iteration(), log.Agent()])


if __name__ == '__main__':
    unittest.main()



from typing import Tuple
import logging
import math

from fastrlapi import core


class _LogCallbackBase(core.AgentCallback):
    """Base class for Callback loggers"""

    def __init__(self, logger: logging.Logger = None, prefix: str = None):
        """Writes all calls to logger with the given prefix.

            Args:
                logger: the logger to log (if None a new logger with level debug is created)
                prefix: a string written in front of each log msg
        """
        self._logger = logger
        if self._logger is None:
            self._logger = logging.getLogger()
        self._prefix = prefix
        if self._prefix is None:
            self._prefix = ''

    def log(self, msg_id: str, *args):
        msg = self._prefix + f'{msg_id:<25}'
        for arg in args:
            if arg is not None:
                msg += str(arg) + ' '
        self._logger.warning(msg)


class _Callbacks(_LogCallbackBase):
    """Logs all AgentCallback calls to a Logger"""

    def __init__(self, logger: logging.Logger = None, prefix: str = None):
        """Writes all calls to a callback function to logger with the given prefix.

        Args:
            logger: the logger to log (if None a new logger with level debug is created)
            prefix: a string written in front of each log msg
            """
        super().__init__(logger=logger, prefix=prefix)

    def on_api_log(self, agent_context: core.AgentContext, api_target: str, log_msg: str):
        msg = f'{api_target:<30}'
        if log_msg:
            msg += ' ' + log_msg
        self.log('on_api_log', msg)

    def on_log(self, agent_context: core.AgentContext, log_msg: str):
        self.log('on_log', log_msg)

    def on_gym_init_begin(self, agent_context: core.AgentContext):
        self.log('on_gym_init_begin', agent_context)

    def on_gym_init_end(self, agent_context: core.AgentContext):
        self.log('on_gym_init_end', agent_context)

    def on_gym_reset_begin(self, agent_context: core.AgentContext, **kwargs):
        self.log('on_gym_reset_begin', agent_context)

    def on_gym_reset_end(self, agent_context: core.AgentContext, reset_result: Tuple, **kwargs):
        self.log('on_gym_reset_end', agent_context)

    def on_gym_step_begin(self, agent_context: core.AgentContext, action):
        self.log('on_gym_step_begin', agent_context)

    def on_gym_step_end(self, agent_context: core.AgentContext, action, step_result: Tuple):
        self.log('on_gym_step_end', agent_context)

    def on_play_begin(self, agent_context: core.AgentContext):
        self.log('on_play_begin', agent_context)

    def on_play_end(self, agent_context: core.AgentContext):
        self.log('on_play_end', agent_context)

    def on_play_episode_begin(self, agent_context: core.AgentContext):
        self.log('on_play_episode_begin', agent_context)

    def on_play_episode_end(self, agent_context: core.AgentContext):
        self.log('on_play_episode_end', agent_context)

    def on_play_step_begin(self, agent_context: core.AgentContext, action):
        self.log('on_play_step_begin', agent_context)

    def on_play_step_end(self, agent_context: core.AgentContext, action, step_result: Tuple):
        self.log('on_play_step_end', agent_context)

    def on_train_begin(self, agent_context: core.AgentContext):
        self.log('on_train_begin', agent_context)

    def on_train_end(self, agent_context: core.AgentContext):
        self.log('on_train_end', agent_context)

    def on_train_iteration_begin(self, agent_context: core.AgentContext):
        self.log('on_train_iteration_begin', agent_context)

    def on_train_iteration_end(self, agent_context: core.AgentContext):
        self.log('on_train_iteration_end', agent_context)


class _AgentContext(_LogCallbackBase):
    """Logs the agent context and its subcontexts after every training iteration / episode played """

    def on_play_episode_end(self, agent_context: core.AgentContext):
        self.log(str(agent_context))

    def on_train_iteration_end(self, agent_context: core.AgentContext):
        self.log(str(agent_context))


class _CallbackCounts(core.AgentCallback):

    def __init__(self):
        self.gym_init_begin_count = 0
        self.gym_init_end_count = 0
        self.gym_reset_begin_count = 0
        self.gym_reset_end_count = 0
        self.gym_step_begin_count = 0
        self.gym_step_end_count = 0
        self.api_log_count = 0
        self.log_count = 0
        self.train_begin_count = 0
        self.train_end_count = 0
        self.train_iteration_begin_count = 0
        self.train_iteration_end_count = 0

    def __str__(self):
        return f'gym_init={self.gym_init_begin_count}:{self.gym_init_end_count} ' + \
               f'gym_reset={self.gym_reset_begin_count}:{self.gym_reset_end_count} ' + \
               f'gym_step={self.gym_step_begin_count}:{self.gym_step_end_count}' + \
               f'train={self.train_begin_count}:{self.train_end_count} ' + \
               f'train_iteration={self.train_iteration_begin_count}:{self.train_iteration_end_count}' + \
               f'api_log={self.api_log_count} log={self.log_count} '

    def on_api_log(self, agent_context: core.AgentContext, api_target: str, log_msg: str):
        self.api_log_count += 1

    def on_log(self, agent_context: core.AgentContext, log_msg: str):
        self.log_count += 1

    def on_gym_init_begin(self, agent_context: core.AgentContext):
        self.gym_init_begin_count += 1

    def on_gym_init_end(self, agent_context: core.AgentContext):
        self.gym_init_end_count += 1

    def on_gym_reset_begin(self, agent_context: core.AgentContext, **kwargs):
        self.gym_reset_begin_count += 1

    def on_gym_reset_end(self, agent_context: core.AgentContext, reset_result: Tuple, **kwargs):
        self.gym_reset_end_count += 1

    def on_gym_step_begin(self, agent_context: core.AgentContext, action):
        self.gym_step_begin_count += 1

    def on_gym_step_end(self, agent_context: core.AgentContext, action, step_result: Tuple):
        self.gym_step_end_count += 1

    def on_train_begin(self, agent_context: core.AgentContext):
        """Called once at the entry of an agent.train() call. """
        self.train_begin_count += 1

    def on_train_end(self, agent_context: core.AgentContext):
        """Called once before exiting an agent.train() call"""
        self.train_end_count += 1

    def on_train_iteration_begin(self, agent_context: core.AgentContext):
        """Called once at the start of a new iteration. """
        self.train_iteration_begin_count += 1

    def on_train_iteration_end(self, agent_context: core.AgentContext):
        """Called once after the current iteration is completed"""
        self.train_iteration_end_count += 1


class Agent(_LogCallbackBase):
    """Logs agent activities to a python logger."""

    def on_api_log(self, agent_context: core.AgentContext, api_target: str, log_msg: str):
        self.log(api_target, log_msg)

    def on_log(self, agent_context: core.AgentContext, log_msg: str):
        self.log(log_msg)


class Duration(_LogCallbackBase):
    """Logs training / play duration definition summary to a logger."""

    def log_duration(self, agent_context: core.AgentContext):
        tc = agent_context.train
        msg = f'num_iterations={tc.num_iterations} '
        if isinstance(tc, core.EpisodesTrainContext):
            ec: core.EpisodesTrainContext = tc
            msg = msg + f'num_episodes_per_iteration={ec.num_episodes_per_iteration} '
        if isinstance(tc, core.StepsTrainContext):
            sc: core.StepsTrainContext = tc
            msg = msg + f'num_steps_per_iteration={sc.num_steps_per_iteration} '
        msg = msg + f'num_max_steps_per_episode={tc.max_steps_per_episode} '
        msg = msg + f'num_iterations_between_plot={tc.num_iterations_between_plot} '
        msg = msg + f'num_iterations_between_eval={tc.num_iterations_between_eval} '
        msg = msg + f'num_episodes_per_eval={tc.num_episodes_per_eval} '
        self.log(f'{"duration":<25}{msg}')

    def on_train_begin(self, agent_context: core.AgentContext):
        self.log_duration(agent_context)


class Iteration(_LogCallbackBase):
    """Logs training iteration summaries to a python logger."""

    def __init__(self, eval_only:bool=False, logger: logging.Logger = None, prefix: str = None):
        """Logs the completion of each training iteration. On an iteration with policy evaluation the
            current average reward/episode and steps/episode is logged as well.

            Args:
                eval_only: if set a log is only created if the policy was re-evaluated in the current iteration.
                logger: the logger to log (if None a new logger with level debug is created)
                prefix: a string written in front of each log msg
        """
        self._eval_only:bool=eval_only
        super().__init__(logger=logger,prefix=prefix)

    def log_iteration(self, agent_context: core.AgentContext):
        tc = agent_context.train
        e = tc.episodes_done_in_training
        if not self._eval_only or (tc.iterations_done_in_training % tc.num_iterations_between_eval == 0):
            msg = f'episodes_done={e:<3} steps_done={tc.steps_done_in_training:<5} '
            if e in tc.loss:
                loss = tc.loss[e]
                if not (isinstance(loss, float) and math.isnan(loss)):
                    msg = msg + f'loss={tc.loss[e]:<7.1f} '
                if isinstance(tc, core.PpoTrainContext):
                    msg = msg + f'[actor={tc.actor_loss[e]:<7.1f} '
                    msg = msg + f'critic={tc.critic_loss[e]:<7.1f}] '
            if e in tc.eval_rewards:
                r = tc.eval_rewards[e]
                msg = msg + f'rewards=({r[0]:.1f},{r[1]:.1f},{r[2]:.1f}) '
            if e in tc.eval_steps:
                s = tc.eval_steps[e]
                msg = msg + f'steps=({s[0]:.1f},{s[1]:.1f},{s[2]:.1f}) '
            prefix = f'iteration {tc.iterations_done_in_training:<2} of {tc.num_iterations} '
            self.log(f'{prefix:<25}{msg}')

    def on_train_iteration_begin(self, agent_context: core.AgentContext):
        tc: core.TrainContext = agent_context.train
        # log the results of a pre-train evaluation (if existing)
        if (0 in tc.eval_rewards) and \
            (tc.episodes_done_in_training == 0) and \
            (tc.iterations_done_in_training == 0):
            self.log_iteration(agent_context)

    def on_train_iteration_end(self, agent_context: core.AgentContext):
        self.log_iteration(agent_context)


class Step(_LogCallbackBase):
    """Logs each environment step to a python logger."""

    def on_gym_step_end(self, agent_context: core.AgentContext, action, step_result: Tuple):
        prefix = ''
        monitor = agent_context.gym._monitor_env
        if monitor:
            prefix = f'[{monitor.gym_env_name} {monitor.instance_id}:{monitor.episodes_done:<3}:' + \
                     f'{monitor.steps_done_in_episode:<3}] '
        tc: core.TrainContext = agent_context.train
        if tc:
            prefix += f'train iteration={tc.iterations_done_in_training:<2} step={tc.steps_done_in_iteration:<4}'
        pc: core.PlayContext = agent_context.play
        if pc:
            prefix += f'play  episode={pc.episodes_done:<2} step={pc.steps_done_in_episode:<5} ' + \
                      f'sum_of_rewards={pc.sum_of_rewards[pc.episodes_done + 1]:<7.1f}'
        (observation, reward, done, info) = step_result
        msg = ''
        if info:
            msg = f' info={msg}'
        self.log(f'{prefix} reward={reward:<5.1f} done={str(done):5} action={action} observation={observation}{msg}')
